# --------------------------------------------------------------------------- #
#   Copyright (c) 2024-2026 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from collections import deque
from contextlib import contextmanager
from functools import cmp_to_key
import random
import sqlite3
from typing import Optional, List

from gremlin_python.process.traversal import Bytecode, Cardinality, Order, P

from .cursors import DummyCursor, GeneratorCursor
from .errors import BadStepArgs, BadStepCombination, UnknownStep
from .filtering import (
    FilterHandler, FilterableElement, FilterQueryBuilder, FilterableQuery, heuristic_filter_split,
)
from .base import (
    GremliteConfig, CanonicalStepNames, ProducerCursor,
    Producer, SingletonProducer, RestartBarrier, PassthroughBarrier, EvaluatorProducer,
    ModulatingProducer, ModulatingEvaluatorProducer, ModulatingElementEvaluatorProducer,
    SideEffect, FilterProducer,
    ElementEvaluatorProducer, ElementFilterProducer,
    VertexConsumerProducer, EdgeConsumerProducer,
    PropertyConsumerProducer, ListConsumerProducer,
    StepStream, Step, TraversalContext
)
import gremlite.querytools as qt
from .querytools import PredicateHandler
from .results import Result


class FilteringProducer(Producer):
    """
    A producer that performs an SQLite query, and may combine *some* subsequent `has()` steps
    into this query, while leaving others to be performed on their own.
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.query_builder = None
        self.initial_query_is_broad = False
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        self.special_build(ss)

        self.c_steps = ss.take_contributing_steps(filter_steps=True)
        filter_steps = self.c_steps.filter_steps

        if self.config.use_basic_heuristic_filtering:
            query_steps, element_steps = self.heuristic_filter_split(filter_steps)
        else:
            query_steps, element_steps = filter_steps, []

        self.query_builder = FilterQueryBuilder(self.config, self.con, query_steps)
        ss.insert_steps(element_steps)

    def make_next_cursor(self):
        fq = self.write_initial_subquery()
        query = self.query_builder.build(fq)
        return qt.run_query(self.con, query, do_close=self.config.read_all_at_once)

    def heuristic_filter_split(self, incoming_steps: List[Step]) -> (List[Step], List[Step]):
        return heuristic_filter_split(incoming_steps, initial_query_is_broad=self.initial_query_is_broad)

    def special_build(self, ss: StepStream):
        """Subclasses may implement, to perform special steps at the start of `build()`. """
        pass  # pragma: no cover

    def write_initial_subquery(self) -> FilterableQuery:
        raise NotImplementedError  # pragma: no cover

    def assemble_outgoing_result(self, next_cursor_item) -> Result:
        """
        :param next_cursor_item: row of length one, holding the signed ID for a graph element,
            i.e. a positive integer for a vertex, negative for an edge.
        :return: Result
        """
        raise NotImplementedError  # pragma: no cover


class V(FilteringProducer):
    """
    V() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.vids = None
        self.instances = None
        super().__init__(con, ss, parent=parent)

    def special_build(self, ss: StepStream):
        self.vids, self.instances = self.base_step.extract_list_of_ids(accept_vertices=True)
        self.initial_query_is_broad = not self.vids

    def write_initial_subquery(self) -> FilterableQuery:
        if self.query_builder.filter_steps:
            query = self.vids or qt.Sign.POS
        else:
            query = qt.write_quads_query(
                sel='s', s=self.vids or None, p=qt.Constants.HAS_LABEL_PREDICATE
            )
        return FilterableQuery(qt.SubjectTypes.VERTICES, query)

    def assemble_outgoing_result(self, next_cursor_item):
        r = self.incoming_result_or_new()

        if self.instances:
            r.note_vertices(self.instances)
            self.instances = None

        vid = next_cursor_item[0]
        r.add_vertex_to_path(vid)
        return r


class E(FilteringProducer):
    """
    E() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.eids = None
        self.instances = None
        super().__init__(con, ss, parent=parent)

    def special_build(self, ss: StepStream):
        self.eids, self.instances = self.base_step.extract_list_of_ids(accept_edges=True)
        self.initial_query_is_broad = not self.eids

    def write_initial_subquery(self) -> FilterableQuery:
        negative_eids = [-eid for eid in self.eids]
        if self.query_builder.filter_steps:
            query = negative_eids or qt.Sign.NEG
        else:
            query = qt.write_quads_query(
                sel='g', g=negative_eids or qt.Sign.NEG
            )
        return FilterableQuery(qt.SubjectTypes.EDGES, query)

    def assemble_outgoing_result(self, next_cursor_item):
        r = self.incoming_result_or_new()

        if self.instances:
            r.note_edges(self.instances)
            self.instances = None

        g = next_cursor_item[0]
        eid = -g
        r.add_edge_to_path(eid)
        return r


class AddV(Producer):
    """
    add_v() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.label = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        args = self.base_step.extract_string_varargs(allowed_numbers=[0, 1], noun='label')
        self.label = args[0] if args else 'vertex'

    def make_next_cursor(self):
        vertex_id = qt.create_vertex(self.con, self.label)
        return DummyCursor([[vertex_id]])

    def assemble_outgoing_result(self, next_cursor_item):
        r = self.incoming_result_or_new()
        vertex_id = next_cursor_item[0]
        r.add_vertex_to_path(vertex_id, label=self.label)
        return r


class AddE(VertexConsumerProducer):
    """
    add_e() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.edge_label = None
        self.source_obj_label = None
        self.target_obj_label = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        args = self.base_step.extract_string_varargs(allowed_numbers=[1], noun='label')
        self.edge_label = args[0]

        self.c_steps = ss.take_contributing_steps(new_edge_steps=True)
        self.source_obj_label = self.c_steps.get_source()
        self.target_obj_label = self.c_steps.get_target()

    def check_incoming_result_type(self, err_msg_extension=''):
        if self.source_obj_label is None or self.target_obj_label is None:
            super().check_incoming_result_type(err_msg_extension=', since not both source and target were given')

        for obj_label in [self.source_obj_label, self.target_obj_label]:
            if obj_label is not None:
                if not self.current_incoming_result.has_labeled_vertex(obj_label):
                    raise BadStepCombination(self.build_err_msg(
                        f'named "{obj_label}" as endpoint, but object cannot be found.'
                    ))

    @property
    def source_vertex(self):
        return (
            self.in_r.last_object if self.source_obj_label is None
            else self.in_r.get_labeled_object(self.source_obj_label)
        )

    @property
    def target_vertex(self):
        return (
            self.in_r.last_object if self.target_obj_label is None
            else self.in_r.get_labeled_object(self.target_obj_label)
        )

    def make_next_cursor(self):
        edge_id = qt.create_edge(
            self.con, self.source_vertex.id, self.edge_label, self.target_vertex.id
        )
        return DummyCursor([[edge_id]])

    def assemble_outgoing_result(self, next_cursor_item):
        r = self.incoming_result_or_new()
        edge_id = next_cursor_item[0]
        r.add_edge_to_path(edge_id, self.source_vertex.id, self.target_vertex.id, label=self.edge_label)
        return r


class As(SideEffect):
    """
    as() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.obj_labels = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        self.obj_labels = self.base_step.extract_string_varargs(nonzero=True, noun='labels')

    def side_effect(self):
        if self.context is TraversalContext.BASIC:
            for label in self.obj_labels:
                self.in_r.label_last_object(label)

    @property
    def do_forward(self):
        if self.context is TraversalContext.PATTERN_MATCHING:
            obj = self.in_r.last_object
            return all(
                obj is self.in_r.get_labeled_object(label)
                for label in self.obj_labels
            )
        else:
            return True


class Count(RestartBarrier):
    """
    count() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.c = 0
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    def for_each_incoming_result(self, r: Result):
        self.c += 1

    @property
    def final_value(self):
        return self.c


class Fold(RestartBarrier):
    """
    fold() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.L = []
        super().__init__(con, ss, parent=parent)
    
    def build(self, ss: StepStream):
        self.base_step.check_no_args()
        
    def for_each_incoming_result(self, r: Result):
        self.L.append(r.last_object)

    @property
    def final_value(self):
        return self.L


class Barrier(PassthroughBarrier):
    """
    barrier() step
    """

    def build(self, ss: StepStream):
        self.base_step.check_no_args()


class Store(ModulatingProducer):
    """
    store() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.list_name = None
        super().__init__(con, ss, parent=parent)

    def special_build(self, ss: StepStream):
        args = self.base_step.extract_string_varargs(allowed_numbers=[1])
        self.list_name = args[0]

    def make_next_cursor(self):
        obj = self.in_r.last_object
        m = self.next_modulator()
        if m is not None:
            with carry_on(self, m) as cur:
                if cur.has_next():
                    obj = cur.fetchone().last_object
                else:
                    obj = None
        r = self.incoming_result_or_new()
        if obj is not None:
            r.add_to_storage_list(self.list_name, obj)
        return DummyCursor([r])


class Cap(RestartBarrier):
    """
    cap() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.list_name = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        args = self.base_step.extract_string_varargs(allowed_numbers=[1])
        self.list_name = args[0]

    @property
    def final_value(self):
        r = self.incoming_result_or_new()
        return r.get_storage_list(self.list_name)


class OrderStep(ModulatingProducer, PassthroughBarrier):
    """
    order() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.orderings = []
        super().__init__(con, ss, parent=parent)

    def special_build(self, ss: StepStream):
        self.base_step.check_no_args()

    def preprocess_by_step(self, step: Step) -> Step:
        ordering = Order.asc
        args = step.args
        if args and isinstance(args[-1], Order):
            ordering = args[-1]
            # We work with a copy of the Step, instead of the original, so that
            # any error messages generated from the Step reflect what the user
            # passed in.
            step = step.copy()
            step.args = args[:-1]
        self.orderings.append(ordering)
        return step

    def compare(self, A: Result, B: Result):
        for a, b, ordering in zip(A.order_keys, B.order_keys, self.orderings):
            if a < b:
                return -1 if ordering is Order.asc else 1
            elif b < a:
                return 1 if ordering is Order.asc else -1
        return 0

    def make_results_list(self):
        # In case there were no `by()` steps, we still need an ordering:
        if len(self.orderings) == 0:
            self.orderings.append(Order.asc)

        # Non-productive sorting modulators are supposed to filter,
        # even if they turn out not to be needed as secondary keys,
        # and even if we are shuffling.
        # Therefore, we start by applying all of them, to all incoming results.
        filtered_results = []
        for r0 in self.all_incoming_results:
            assert isinstance(r0, Result)
            r0.clear_order_keys()
            for modulator in self.modulators:
                if modulator is None:
                    key = r0.last_object_sort_key
                else:
                    with carry_on(self, modulator, from_result=r0) as cur:
                        if cur.has_next():
                            r1 = cur.fetchone()
                            key = r1.last_object_sort_key
                        else:
                            break
                r0.add_order_key(key)
            else:
                # We never broke, i.e. every modulator was productive.
                filtered_results.append(r0)

        if Order.shuffle in self.orderings:
            # We implement the rule that, "If anything's a shuffle, it's all a shuffle."
            random.shuffle(filtered_results)
        else:
            # Otherwise, there is no shuffle. Everything is either ascending or descending.
            filtered_results.sort(
                key=cmp_to_key(lambda A, B: self.compare(A, B))
            )

        return filtered_results


class Unfold(ListConsumerProducer):
    """
    unfold() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    def make_next_cursor(self):
        return DummyCursor(self.in_r.last_object)

    def assemble_outgoing_result(self, next_cursor_item) -> Result:
        self.in_r.move_to_value(next_cursor_item)
        return self.in_r


class Key(EvaluatorProducer, PropertyConsumerProducer):
    """
    key() step
    """

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    def evaluate(self, r: Result):
        return r.last_object.key


class Value(EvaluatorProducer, PropertyConsumerProducer):
    """
    value() step
    """

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    def evaluate(self, r: Result):
        return r.last_object.value


class ValueMap(ModulatingProducer, ElementEvaluatorProducer):
    """
    value_map() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.prop_names = None
        self.incl_tokens = False
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        base_step = self.base_step

        args = base_step.args
        if args and isinstance(args[0], bool):
            self.incl_tokens = args[0]
            # We work with a copy of the Step, instead of the original, so that
            # any error messages generated from the Step reflect what the user
            # passed in.
            base_step = base_step.copy()
            base_step.args = args[1:]

        args = base_step.extract_string_varargs(noun='property names')
        self.prop_names = args or None

        ModulatingProducer.build(self, ss)

    def evaluate(self, r: Result):
        obj = r.last_object

        d0 = r.build_element_info_dict(
            obj, incl_tokens=self.incl_tokens, prop_names=self.prop_names
        )

        d1 = {}
        for k, v in d0.items():
            modulator = self.next_modulator()
            if modulator is not None:
                r0 = Result(self.con, self.ss)
                r0.move_to_value(v)
                with carry_on(self, modulator, from_result=r0) as cur:
                    if cur.has_next():
                        r1 = cur.fetchone()
                        v = r1.last_object
                    else:
                        continue
            d1[k] = v

        return d1


class ElementMap(ElementEvaluatorProducer):
    """
    element_map() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.prop_names = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        args = self.base_step.extract_string_varargs(noun='property names')
        self.prop_names = args or None

    def evaluate(self, r: Result):
        obj = r.last_object
        return r.build_element_info_dict(
            obj, incl_tokens=True, incl_endpts=True,
            prop_names=self.prop_names, single_props=True
        )


class Constant(ElementEvaluatorProducer):
    """
    constant() step
    """

    def build(self, ss: StepStream):
        if len(self.base_step.args) != 1:
            raise BadStepArgs(self.build_err_msg("Pass exactly one arg."))
        a = self.base_step.args[0]
        # Any constraints on the type of the constant?
        # For now, we'll just say it can't be Bytecode.
        if isinstance(a, Bytecode):
            raise NotImplementedError

    def evaluate(self, r: Result):
        return self.base_step.args[0]


class Id(ElementEvaluatorProducer):
    """
    id() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        super().__init__(con, ss, parent=parent)
        self.accept_vertex_properties = True

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    def evaluate(self, r: Result):
        return r.last_object.id


class Label(ElementEvaluatorProducer):
    """
    label() step
    """

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    def evaluate(self, r: Result):
        obj = r.last_object
        r.determine_label(obj)
        return obj.label


class NoneStep(Producer):
    """
    none() step
    """

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    def make_next_cursor(self):
        return DummyCursor.emtpy_cursor()


class Identity(FilterProducer):
    """
    identity() step
    """

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    def passes(self, r: Result):
        return True


class SimplePath(FilterProducer):
    """
    simple_path()
    """

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    def passes(self, r: Result):
        return r.path_is_simple()


class Not(FilterProducer):
    """
    not() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.bytecode = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        args = self.base_step.extract_bytecode_varargs(allowed_numbers=[1])
        self.bytecode = args[0]

    def passes(self, r: Result):
        with carry_on(self, self.bytecode) as cur:
            return not cur.has_next()


class And(FilterProducer):
    """
    and() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.bytecodes = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        self.bytecodes = self.base_step.extract_bytecode_varargs(nonzero=True)

    def passes(self, r: Result):
        for bc in self.bytecodes:
            with carry_on(self, bc) as cur:
                if not cur.has_next():
                    return False
        return True


class Where(FilterProducer):
    """
    where() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.bytecode = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        self.bytecode = self.base_step.extract_bytecode_varargs(allowed_numbers=[1])[0]

    def passes(self, r: Result):
        with carry_on(self, self.bytecode, context=TraversalContext.PATTERN_MATCHING) as cur:
            return cur.has_next()


class Or(FilterProducer):
    """
    or() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.bytecodes = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        self.bytecodes = self.base_step.extract_bytecode_varargs(nonzero=True)

    def passes(self, r: Result):
        for bc in self.bytecodes:
            with carry_on(self, bc) as cur:
                if cur.has_next():
                    return True
        return False


class Limit(FilterProducer):
    """
    limit() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.counter = 0
        self.limit = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        args = self.base_step.extract_varargs_of_types([int], allowed_numbers=[1])
        self.limit = args[0]

    def passes(self, r: Result):
        self.counter += 1
        # `limit()` is not a pure filter step; when the limit is reached, it has
        # to forcefully stop everything by raising `StopIteration`. (If it just
        # returned an empty cursor, we'd continue trying again with a new incoming
        # result.)
        if self.counter <= self.limit:
            return True
        else:
            raise StopIteration


class Has(ElementFilterProducer, FilterHandler, PredicateHandler):
    """
    has() step

    This handles the case in which a `has()` step operates standalone.
    Other producers may consume some `has()` steps as contributing steps, while
    leaving others to be handled by this class.
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.given_strings = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        pass

    def passes(self, r: Result):
        elt0 = FilterableElement(r)
        elt1 = self.dispatch(self.base_step, elt0)
        return elt1 is elt0

    def has_label(self, filter_step: Step, labels, filterable: FilterableElement) -> FilterableElement:
        r = filterable.result
        obj = r.last_object
        r.determine_label(obj)

        if labels and isinstance(labels[0], P):
            pred = labels[0]
            self.given_strings = [obj.label]
            passes = self.handle_predicate(pred)
        else:
            passes = obj.label in labels

        return filterable if passes else FilterableElement(None)

    def unary_has(self, filter_step: Step, key: str, filterable: FilterableElement) -> FilterableElement:
        r = filterable.result
        obj = r.last_object
        d = r.get_properties(obj, prop_names=[key], single=True, value_only=True)
        return filterable if d else FilterableElement(None)

    def binary_has(self, filter_step: Step, key: str, value, filterable: FilterableElement) -> FilterableElement:
        r = filterable.result
        obj = r.last_object
        values = r.get_properties(obj, prop_names=[key], value_only=True, as_list=True)

        if isinstance(value, P):
            pred = value
            self.given_strings = values
            passes = self.handle_predicate(pred)
        else:
            passes = value in values

        return filterable if passes else FilterableElement(None)

    def starting_with(self, value):
        return any(s.startswith(value) for s in self.given_strings)

    def containing(self, value):
        return any(s.find(value) >= 0 for s in self.given_strings)

    def ending_with(self, value):
        return any(s.endswith(value) for s in self.given_strings)

    def gt(self, value):
        return any(s > value for s in self.given_strings)

    def lt(self, value):
        return any(s < value for s in self.given_strings)

    def gte(self, value):
        return any(s >= value for s in self.given_strings)

    def lte(self, value):
        return any(s <= value for s in self.given_strings)

    def within(self, value):
        return any(s in value for s in self.given_strings)


"""
has_label() step
"""
HasLabel = Has


class IncidentEdgeTraverser(VertexConsumerProducer, FilteringProducer):
    """
    Abstract base class for steps that move from a vertex either to
    an incident edge, or across an incident edge to an adjacent vertex.
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.specific_labels = None
        self.p = None
        self.starting_vertex_id = None
        super().__init__(con, ss, parent=parent)

    def special_build(self, ss: StepStream):
        self.specific_labels = self.base_step.extract_string_varargs(noun='edge labels')
        self.p = (
            qt.write_strings_to_ids_query(self.specific_labels, id_valence=-1)
            if self.specific_labels else qt.Sign.NEG
        )

    @property
    def sel(self):
        raise NotImplementedError  # pragma: no cover

    @property
    def s(self):
        raise NotImplementedError  # pragma: no cover

    @property
    def o(self):
        raise NotImplementedError  # pragma: no cover

    @property
    def subject_type(self):
        return qt.SubjectTypes.EDGES if 'g' in self.sel else qt.SubjectTypes.VERTICES

    def write_initial_subquery(self) -> FilterableQuery:
        self.starting_vertex_id = self.in_r.last_object.id

        base_queries = []
        for sel, s, o in zip(self.sel, self.s, self.o):
            base_queries.append(qt.write_quads_query(sel=sel, s=s, p=self.p, o=o))
        query = qt.query_join(' UNION ALL ', base_queries)

        return FilterableQuery(self.subject_type, query)

    def assemble_outgoing_result(self, next_cursor_item):
        r = self.incoming_result_or_new()
        signed_id = next_cursor_item[0]
        r.add_object_to_path(signed_id)
        return r


class BothE(IncidentEdgeTraverser):
    """
    both_e() step. Move from the current vertex to both its incoming and its outgoing edges.
    """

    @property
    def sel(self):
        return ['g', 'g']

    @property
    def s(self):
        return [None, self.starting_vertex_id]

    @property
    def o(self):
        return [self.starting_vertex_id, None]


class InE(IncidentEdgeTraverser):
    """
    in_e() step. Move from the current vertex to its incoming edges.
    """

    @property
    def sel(self):
        return ['g']

    @property
    def s(self):
        return [None]

    @property
    def o(self):
        return [self.starting_vertex_id]


class OutE(IncidentEdgeTraverser):
    """
    out_e() step. Move from the current vertex to its outgoing edges.
    """

    @property
    def sel(self):
        return ['g']

    @property
    def s(self):
        return [self.starting_vertex_id]

    @property
    def o(self):
        return [None]


class Both(IncidentEdgeTraverser):
    """
    both() step. Hop from the current vertex to neighboring vertices, along both incoming
    and outgoing edges.
    """

    @property
    def sel(self):
        return ['s', 'o']

    @property
    def s(self):
        return [None, self.starting_vertex_id]

    @property
    def o(self):
        return [self.starting_vertex_id, None]


class In(IncidentEdgeTraverser):
    """
    in() step. Hop from the current vertex to neighboring vertices, along incoming edges.
    """

    @property
    def sel(self):
        return ['s']

    @property
    def s(self):
        return [None]

    @property
    def o(self):
        return [self.starting_vertex_id]


class Out(IncidentEdgeTraverser):
    """
    out() step. Hop from the current vertex to neighboring vertices, along outgoing edges.
    """

    @property
    def sel(self):
        return ['o']

    @property
    def s(self):
        return [self.starting_vertex_id]

    @property
    def o(self):
        return [None]


class EdgeToEndpointTraverser(EdgeConsumerProducer):
    """
    Abstract base class for the steps out_v(), in_v(), other_v(), both_v(), that move
    from an edge to one or both of its endpoint vertices.
    """

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    def make_next_cursor(self):
        ids = [v.id for v in self.goal_vertices]
        return DummyCursor(ids)

    @property
    def goal_vertices(self):
        """
        Subclasses must return the list of vertices we want to move to.
        """
        raise NotImplementedError  # pragma: no cover

    def assemble_outgoing_result(self, next_cursor_item):
        r = self.incoming_result_or_new()
        # Because of the way Edge objects are formed in the Result object,
        # the Edge will always already have references to its two endpoint vertices.
        vertex_id = next_cursor_item
        r.add_vertex_to_path(vertex_id)
        return r


class InV(EdgeToEndpointTraverser):
    """
    in_v() step (move from an edge to its target vertex)
    """

    @property
    def goal_vertices(self):
        return [self.in_r.last_object.inV]


class OutV(EdgeToEndpointTraverser):
    """
    out_v() step (move from an edge to its source vertex)
    """

    @property
    def goal_vertices(self):
        return [self.in_r.last_object.outV]


class BothV(EdgeToEndpointTraverser):
    """
    both_v() step (move from an edge to both of its adjacent
    vertices)
    """

    @property
    def goal_vertices(self):
        edge = self.in_r.last_object
        return [edge.outV, edge.inV]


class OtherV(EdgeToEndpointTraverser):
    """
    other_v() step (move from an edge to that one of its endpoints
    that we did *not* just visit
    """

    @property
    def goal_vertices(self):
        edge = self.in_r.last_object
        origin_vertex = self.in_r.penultimate_object
        both_vertices = [edge.outV, edge.inV]
        if origin_vertex not in both_vertices:
            msg = "Previous object not one of edge's endpoints."
            raise BadStepCombination(self.build_err_msg(msg))
        both_vertices.remove(origin_vertex)
        other_vertex = both_vertices.pop()
        return [other_vertex]


class Property(SideEffect):
    """
    property() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.card = Cardinality.single
        self.prop_name = None
        self.prop_value = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        args = self.base_step.args

        if len(args) == 3:
            c = args[0]
            if not isinstance(c, Cardinality):
                raise BadStepArgs(self.build_err_msg("When passing 3 args, first should be a Cardinality"))
            self.card = c
            args = args[1:]

        if not len(args) == 2:
            raise BadStepArgs(self.build_err_msg("Pass either (Cardinality, key, value) or (key, value)."))

        prop_name, prop_value = args

        if not isinstance(prop_name, str):
            raise BadStepArgs(self.build_err_msg("property name should be string."))

        if not isinstance(prop_value, (int, float, bool, str, type(None))):
            raise BadStepArgs(self.build_err_msg("property value should be int, float, bool, str, or None"))

        self.prop_name = prop_name
        self.prop_value = prop_value

    def side_effect(self):
        r = self.in_r
        r.set_property(r.last_object, self.prop_name, self.prop_value, cardinality=self.card)


class Properties(ElementEvaluatorProducer):
    """
    properties() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.specific_names = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        self.specific_names = self.base_step.extract_string_varargs(noun='property names') or None

    @property
    def is_multi(self):
        return True

    def evaluate(self, r: Result):
        return r.get_properties(r.last_object, prop_names=self.specific_names, as_list=True)


class Values(Properties):
    """
    values() step
    """

    def evaluate(self, r: Result):
        props = super().evaluate(r)
        return [p.value for p in props]


class Coalesce(Producer):
    """
    coalesce() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.bytecodes = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        self.bytecodes = self.base_step.extract_bytecode_varargs()

    def make_next_cursor(self):
        return GeneratorCursor(self.follow_first_successful_traversal())

    def follow_first_successful_traversal(self):
        """
        Follow the first traversal (if any) that produces one or more results.

        :return: generator
        """
        for bc in self.bytecodes:
            with carry_on(self, bc) as cur:
                if cur.has_next():
                    while cur.has_next():
                        yield cur.fetchone()
                    break


class Union(Producer):
    """
    union() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.bytecodes = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        self.bytecodes = self.base_step.extract_bytecode_varargs()

    def make_next_cursor(self):
        return GeneratorCursor(self.follow_all_traversals())

    def follow_all_traversals(self):
        """
        Follow each traversal on the incoming result, and yield results.

        :return: generator
        """
        for bc in self.bytecodes:
            with carry_on(self, bc) as cur:
                while cur.has_next():
                    yield cur.fetchone()


class FlatMap(Union):
    """
    flat_map() step
    """

    def build(self, ss: StepStream):
        self.bytecodes = self.base_step.extract_bytecode_varargs(allowed_numbers=[1])


class Select(ModulatingEvaluatorProducer):
    """
    select() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.obj_labels = None
        super().__init__(con, ss, parent=parent)

    def special_build(self, ss: StepStream):
        self.obj_labels = self.base_step.extract_string_varargs(nonzero=True, noun='step labels')

    def evaluate(self, r: Result):
        r0 = r
        d = {}

        for obj_label in self.obj_labels:
            obj = r0.get_labeled_object(obj_label)

            # In TinkerGraph, if you try to select a label that doesn't exist,
            # it's not an error; you just get no result.
            if obj is None:
                raise StopIteration

            modulator = self.next_modulator()

            if modulator is not None:
                assert isinstance(modulator, Bytecode)
                # Follow the traversal, and accept the first result.
                # If there are no results, then the `select()` step returns no results.
                # Question: could we make `mixed_copy()` work here instead of full `copy()`?
                r = r0.copy(stop_at_latest=obj_label)
                with carry_on(self, modulator, from_result=r) as cur:
                    if cur.has_next():
                        result = cur.fetchone()
                        obj = result.last_object
                    else:
                        raise StopIteration

            d[obj_label] = obj

        # If only one thing was selected, return it as a bare value.
        # Otherwise, we return a whole dictionary.
        if len(d) == 1:
            _, d = d.popitem()

        return d


class Path(ModulatingProducer, EvaluatorProducer):
    """
    path() step
    """

    def special_build(self, ss: StepStream):
        self.base_step.check_no_args()

    def evaluate(self, r: Result):
        r0 = r
        N = len(r0)

        objects = []
        for k in range(N):
            modulator = self.next_modulator()
            if modulator is None:
                obj = r0.get_object_by_index(k)
            else:
                # Question: could we make `mixed_copy()` work here instead of full `copy()`?
                r = r0.copy(rewind=N-1-k)
                with carry_on(self, modulator, from_result=r) as cur:
                    if cur.has_next():
                        obj = cur.fetchone().last_object
                    else:
                        # In Tinkergraph, if any modulator is non-productive,
                        # then we get no path at all.
                        raise StopIteration
            objects.append(obj)

        return r0.make_path(objects)


class Project(ModulatingElementEvaluatorProducer):
    """
    project() step
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.dict_keys = None
        super().__init__(con, ss, parent=parent)

    def special_build(self, ss: StepStream):
        self.dict_keys = self.base_step.extract_string_varargs(nonzero=True, noun='dictionary keys')

    def evaluate(self, r: Result):
        d = {}

        for k in self.dict_keys:
            modulator = self.next_modulator()
            if modulator is None:
                obj = self.in_r.last_object
            else:
                with carry_on(self, modulator) as cur:
                    if cur.has_next():
                        obj = cur.fetchone().last_object
                    else:
                        obj = None
            if obj is not None:
                d[k] = obj

        return d


class RepeatQueueItem:
    """
    Each item in the BFS queue for a `repeat()` step records the Result that was reached, and the
    number of applications of our main traversal that it took to get there.
    """

    def __init__(self, result: Result, num_apps: int):
        self.result = result
        self.num_apps = num_apps


class RepeatEmitUntil(Producer):
    """
    repeat() step, with optional emit(), until(), and times() modifiers.
    """

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.do_emit = False
        self.main_trav = None
        self.rep_limit = None
        self.until_trav = None

        self.emit_incoming = False
        self.until_incoming = False

        self.bfs_queue = deque()

        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        ss.rewind()
        self.c_steps = ss.take_contributing_steps(allow_repeats=False, repeater_steps=True)
        all_steps = self.c_steps.repeater_steps

        post_repeat = False

        for step in all_steps:
            if step.name == CanonicalStepNames.emit:
                step.check_no_args()
                self.do_emit = True
                if not post_repeat:
                    self.emit_incoming = True
            elif step.name == CanonicalStepNames.repeat:
                args = step.extract_bytecode_varargs(allowed_numbers=[1])
                self.main_trav = args[0]
                post_repeat = True
            elif step.name == CanonicalStepNames.times:
                args = step.extract_varargs_of_types([int], allowed_numbers=[1])
                self.rep_limit = args[0]
            elif step.name == CanonicalStepNames.until:
                args = step.extract_bytecode_varargs(allowed_numbers=[1])
                self.until_trav = args[0]
                if not post_repeat:
                    self.until_incoming = True

        if not post_repeat:
            raise BadStepCombination(self.build_err_msg("Missing `repeat()` step."))

    def enqueue(self, result: Result, num_apps: int):
        self.bfs_queue.append(RepeatQueueItem(result, num_apps))

    def make_next_cursor(self):
        self.bfs_queue.clear()
        self.enqueue(self.in_r.mixed_copy(), 0)
        cur = GeneratorCursor(self.bfs())
        return cur

    def bfs(self):
        """
        This method is designed to return a generator when invoked. So it uses `yield`
        to return results, and may return `None` to indicate that there are no more
        results to be generated.

        :return: generator
        """
        while self.bfs_queue:
            q = self.bfs_queue.popleft()
            r0 = q.result
            apps = q.num_apps

            do_emit = (apps == 0 and self.emit_incoming) or (apps > 0 and self.do_emit)
            do_check_until = (apps == 0 and self.until_incoming) or (apps > 0 and self.until_trav is not None)
            met_until_check = False

            if do_check_until:
                with carry_on(self, self.until_trav, from_result=r0) as cur:
                    if cur.has_next():
                        met_until_check = True

            if do_emit or met_until_check:
                yield r0

            if not met_until_check and (self.rep_limit is None or apps < self.rep_limit):
                with carry_on(self, self.main_trav, from_result=r0) as cur:
                    n = apps + 1
                    for r in cur.producer:
                        r1 = r.mixed_copy()
                        self.enqueue(r1, n)


class Drop(SideEffect):

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        super().__init__(con, ss, parent=parent)
        self.accept_vertex_properties = True
        self.accept_edge_properties = True

    def build(self, ss: StepStream):
        self.base_step.check_no_args()

    @property
    def do_forward(self):
        return False

    def side_effect(self):
        r = self.in_r
        if self.incoming_result_id_valence == 1:
            qt.completely_remove_vertex(self.con, r.last_object.id)
        elif self.incoming_result_id_valence == -1:
            qt.completely_remove_edges(self.con, -r.last_object.id)
        elif self.incoming_subject_type == qt.SubjectTypes.VERTICES:
            qt.completely_remove_vertex_property(self.con, r.last_object.id)
        elif self.incoming_subject_type == qt.SubjectTypes.EDGES:
            prop = r.last_object
            qt.completely_remove_edge_property(self.con, prop.element.id, prop.key)


class SideEffectStep(SideEffect):

    def __init__(self, con: sqlite3.Connection, ss: StepStream, parent=None):
        self.bytecode = None
        super().__init__(con, ss, parent=parent)

    def build(self, ss: StepStream):
        args = self.base_step.extract_bytecode_varargs(allowed_numbers=[1])
        self.bytecode = args[0]

    def side_effect(self):
        with carry_on(self, self.bytecode) as cur:
            if cur.has_next():
                cur.fetchone()


################################################################################

# This is a lookup from step names to producer classes.
# Only those step names are listed that are *initial producer steps*,
# i.e. ones that can form a producer (and may consume subsequent steps).
# For example, `repeat` is an initial producer step, but `times` is not,
# because it can only contribute to a `repeat` that has already been encountered.
all_known_initial_producer_steps = {
    CanonicalStepNames.V: V,
    CanonicalStepNames.E: E,
    CanonicalStepNames.addE: AddE,
    CanonicalStepNames.addV: AddV,
    CanonicalStepNames.and_: And,
    CanonicalStepNames.as_: As,
    CanonicalStepNames.barrier: Barrier,
    CanonicalStepNames.both: Both,
    CanonicalStepNames.bothE: BothE,
    CanonicalStepNames.bothV: BothV,
    CanonicalStepNames.cap: Cap,
    CanonicalStepNames.coalesce: Coalesce,
    CanonicalStepNames.constant: Constant,
    CanonicalStepNames.count: Count,
    CanonicalStepNames.drop: Drop,
    CanonicalStepNames.elementMap: ElementMap,
    CanonicalStepNames.emit: RepeatEmitUntil,
    CanonicalStepNames.flatMap: FlatMap,
    CanonicalStepNames.fold: Fold,
    CanonicalStepNames.has: Has,
    CanonicalStepNames.hasLabel: HasLabel,
    CanonicalStepNames.id_: Id,
    CanonicalStepNames.identity: Identity,
    CanonicalStepNames.in_: In,
    CanonicalStepNames.inE: InE,
    CanonicalStepNames.inV: InV,
    CanonicalStepNames.key: Key,
    CanonicalStepNames.label: Label,
    CanonicalStepNames.limit: Limit,
    CanonicalStepNames.none: NoneStep,
    CanonicalStepNames.not_: Not,
    CanonicalStepNames.or_: Or,
    CanonicalStepNames.order: OrderStep,
    CanonicalStepNames.otherV: OtherV,
    CanonicalStepNames.out: Out,
    CanonicalStepNames.outE: OutE,
    CanonicalStepNames.outV: OutV,
    CanonicalStepNames.path: Path,
    CanonicalStepNames.project: Project,
    CanonicalStepNames.properties: Properties,
    CanonicalStepNames.property_: Property,
    CanonicalStepNames.repeat: RepeatEmitUntil,
    CanonicalStepNames.select: Select,
    CanonicalStepNames.sideEffect: SideEffectStep,
    CanonicalStepNames.simplePath: SimplePath,
    CanonicalStepNames.store: Store,
    CanonicalStepNames.unfold: Unfold,
    CanonicalStepNames.union: Union,
    CanonicalStepNames.until: RepeatEmitUntil,
    CanonicalStepNames.value: Value,
    CanonicalStepNames.valueMap: ValueMap,
    CanonicalStepNames.values: Values,
    CanonicalStepNames.where: Where,
}


def bytecode_to_producer_chain(config: GremliteConfig, con: sqlite3.Connection, bytecode: Bytecode,
                               seed_producer: Optional[Producer] = None) -> Producer:
    """
    Scan a bytecode representation of a Gremlin traversal, chunk it into
    Producers, linked into a chain, and return the final Producer in the chain.
    """
    last_producer = seed_producer
    ss = StepStream(config, bytecode, existing_stream=seed_producer.ss if seed_producer else None)
    while ss.has_next():
        next_step = ss.peek_next()
        builder = all_known_initial_producer_steps.get(next_step.name)
        if builder is None:
            raise UnknownStep(next_step.name)  # pragma: no cover
        last_producer = builder(con, ss, parent=last_producer)
    return last_producer


@contextmanager
def carry_on(producer: Producer, bytecode: Bytecode, from_result=None, context=None):
    """
    Apply an ongoing traversal.

    :param producer: the Producer that wants to apply the ongoing traversal.
    :param bytecode: Bytecode representing the ongoing traversal.
    :param from_result: optional incoming Result. If not given, we use
        the producer's incoming result, or new.
    :param context: optional TraversalContext for the producer chain.
    :return: ProducerCursor
    """
    r = from_result or producer.incoming_result_or_new()
    seed = SingletonProducer(r, context=context)
    pro = bytecode_to_producer_chain(producer.config, producer.con, bytecode, seed_producer=seed)
    cur = ProducerCursor(pro)
    yield cur
    if not cur.stopped_iteration:
        # If and only if the cursor wasn't actually fully exhausted, we need to
        # do an abort here. When the cursor was *not* fully exhausted, this ensures
        # that any calls to `push_state()` on the `Result` object get their
        # corresponding `pop_state()`. If the cursor *was* fully exhausted, then
        # we must *not* call `abort()`, or else we'll cause *extra* calls to `pop_state()`,
        # which of course we do not want.
        pro.abort()



# --------------------------------------------------------------------------- #
#   Copyright (c) 2024-2026 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

import sqlite3
from typing import List, Optional

from gremlin_python.process.traversal import P

from .base import Step, CanonicalStepNames, GremliteConfig
from .errors import BadStepArgs, BadStepCombination
from .results import Result
import gremlite.querytools as qt


def heuristic_filter_split(
        incoming_steps: List[Step], initial_query_is_broad=False) -> (List[Step], List[Step]):
    """
    Divide a list of `has()` steps into those that should be applied on the query side,
    and those that should be applied on the element side.

    Our strategy is simple: In almost all cases, steps involving text predicates are kept
    on the element side, while all others are put on the query side.

    The reason for this is that text predicate filters are thought to be potentially very costly
    to apply on the query side. Consider for example the following step:

        has('name', TextP.lt('Foo'))

    In order to apply this filter on the query side, we have to constrain the `o` column to lie
    within the set of all integer values that happen to encode strings that are lexicographically
    less than 'Foo'. This could be a huge set, and it is not an interval but a lawless set of
    scattered integers (because strings are added to the strings table not in lexicographic order,
    but in the order they first appear in the database). Depending on the size of the database,
    this could potentially be a costly query to carry out.

    Compare this to what it means to instead apply a text predicate "on the element side." This
    means that, if say we are filtering vertices, we instead grab the actual value of the 'name'
    property for each incoming vertex, and check whether that value is < 'Foo'. If there is a
    small number of incoming vertices, this seems like a far better approach.

    The one exception is that if the initial query is thought to be "broad" (see below), *and*
    if *all* of the given filter steps involve text predicates, then the first filter step is put
    on the query side, while all others remain on the element side.

    A "broad" initial query is one that is thought to do such a poor job of narrowing down,
    that we are actually better off applying a text predicate on the query side than not.
    For example, in `V(2, 3, 5)` the initial query selects exactly three vertices, and that
    is a very narrow initial query; but in `V()` the initial query is that `s > 0` which is
    very broad.
    """
    if not incoming_steps:
        return [], []

    non_text_pred_steps, text_pred_steps = [], []
    for step in incoming_steps:
        if any(qt.is_text_predicate(a) for a in step.args):
            text_pred_steps.append(step)
        else:
            non_text_pred_steps.append(step)

    if initial_query_is_broad and not non_text_pred_steps:
        query_steps, element_steps = text_pred_steps[:1], text_pred_steps[1:]
    else:
        query_steps, element_steps = non_text_pred_steps, text_pred_steps

    return query_steps, element_steps


class Filterable:
    """
    Represents a thing to be filtered.
    """
    ...


class FilterableQuery(Filterable):

    def __init__(self, subject_type: qt.SubjectTypes, s_subquery: str):
        """
        :param subject_type: tells what type of subject we are filtering
        :param s_subquery: optional subquery on the s column. Can be anything that
            could be passed to the `s` kwarg of the `querytools.write_quads_query()` function.
        """
        self.subject_type = subject_type
        self.s_subquery = s_subquery


class FilterableElement(Filterable):

    def __init__(self, result: Optional[Result]):
        """
        :param result: a Result, or None
        """
        self.result = result


class FilterHandler:

    def dispatch(self, filter_step: Step, filterable: Filterable) -> Filterable:
        args = filter_step.args

        if filter_step.name == CanonicalStepNames.has:
            n = len(args)
            if n == 1:
                # has(key)
                key = args[0]
                filter_step.check_arg(key, [str], subject='key arg')
                return self.unary_has(filter_step, key, filterable)

            elif n == 2:
                # has(key, value)
                key, value = args

                filter_step.check_arg(key, [str], subject='key arg')
                filter_step.check_property_value_arg(value, subject='value arg')

                return self.binary_has(filter_step, key, value, filterable)

            elif n == 3:
                # has(label, key, value)
                label, key, value = args

                filter_step.check_arg(label, [str], subject='label arg')
                filter_step.check_arg(key, [str], subject='key arg')
                filter_step.check_property_value_arg(value, subject='value arg')

                filterable = self.dispatch(
                    Step(f'{filter_step.loc} (part 1)', [CanonicalStepNames.hasLabel, label]),
                    filterable
                )
                return self.dispatch(
                    Step(f'{filter_step.loc} (part 2)', [CanonicalStepNames.has, key, value]),
                    filterable
                )

            else:
                raise BadStepArgs(filter_step.err_msg('Wrong number of args'))

        elif filter_step.name == CanonicalStepNames.hasLabel:
            # has_label(*labels)
            labels = args

            if len(labels) == 1:
                filter_step.check_arg(labels[0], [str, P], subject='Label arg')
            else:
                filter_step.extract_string_varargs(nonzero=True)

            return self.has_label(filter_step, labels, filterable)

        else:
            raise BadStepCombination(filter_step.err_msg('Unknown filter step'))  # pragma: no cover

    def unary_has(self, filter_step: Step, key: str, filterable: Filterable) -> Filterable:
        raise NotImplementedError  # pragma: no cover

    def binary_has(self, filter_step: Step, key: str, value, filterable: Filterable) -> Filterable:
        raise NotImplementedError  # pragma: no cover

    def has_label(self, filter_step: Step, labels, filterable: Filterable) -> Filterable:
        raise NotImplementedError  # pragma: no cover


class KnownEmptyResult(Exception):
    """Query is guaranteed to be empty."""
    ...


class FilterQueryBuilder(FilterHandler):
    """
    Builds an SQLite query string, based on a given list of `has()` steps.
    """

    def __init__(self, config: GremliteConfig, con: sqlite3.Connection, filter_steps: List[Step]):
        """
        :param con: connection
        :param filter_steps: list of `has()` steps
        """
        self.config = config
        self.con = con
        self.filter_steps = filter_steps

    def build(self, filterable: FilterableQuery) -> str:
        """Build the SQLite query string. """
        for step in self.filter_steps:
            try:
                filterable = self.dispatch(step, filterable)
            except KnownEmptyResult:
                return qt.write_quads_query(sel='s', s=qt.Constants.IMPOSSIBLE_SUBJECT)
        return filterable.s_subquery

    def unary_has(self, filter_step: Step, key: str, filterable: FilterableQuery) -> FilterableQuery:
        subject_type = filterable.subject_type
        s_subquery = filterable.s_subquery

        s = None
        p = qt.encode_property_name(self.con, key, merge=False)
        g = None

        if p is None:
            # This means the property name doesn't exist yet at all, so there are certainly no quads that have it.
            raise KnownEmptyResult
        else:
            # Since o is unconstrained, we cannot use our (ops) index. Therefore we
            # have to be careful not to formulate a [p|s] constraint pattern, since we have no (ps) index.
            # If s is exact, we can use our (sp) index, but if it is not, then we must
            # use our (pg) index instead.
            if self.config.use_p_bar_s_pattern:
                # This is purely for testing purposes, to verify that our testing framework does indeed raise
                # exceptions on unsupported patterns.
                s = s_subquery
            elif s_subquery is None or isinstance(s_subquery, qt.Sign):
                # The s_subquery is inexact, but we can incorporate the subject type using
                # the g column instead. We'll have a [pg|] pattern for edge properties, and [p|g] for
                # vertex properties, and in either case will use index (pg).
                g = qt.Constants.NON_ID_G_VALUE if subject_type is qt.SubjectTypes.EDGES else qt.Sign.POS
            else:
                # The s_subquery is exact, so we have an [sp|] pattern and will use index (sp).
                s = s_subquery

        # We select 'DISTINCT s' in order to filter out the repeats that would otherwise arise if the
        # requested property p is present in plural cardinality.
        query = qt.write_quads_query(sel='DISTINCT s', s=s, p=p, g=g)
        return FilterableQuery(subject_type, query)

    def binary_has(self, filter_step: Step, key: str, value, filterable: FilterableQuery) -> FilterableQuery:
        subject_type = filterable.subject_type
        s_subquery = filterable.s_subquery

        p = qt.encode_property_name(self.con, key, merge=False)
        o = qt.encode_property_value(self.con, value, merge=False)

        if p is None or o is None:
            # This means the property name or value doesn't exist yet at all,
            # so there are certainly no quads that have this combination.
            raise KnownEmptyResult
        else:
            s = qt.Sign.valence(subject_type) if s_subquery is None else s_subquery

        query = qt.write_quads_query(sel='s', s=s, p=p, o=o)
        return FilterableQuery(subject_type, query)

    def has_label(self, filter_step: Step, labels, filterable: FilterableQuery) -> FilterableQuery:
        subject_type = filterable.subject_type
        s_subquery = filterable.s_subquery

        label_constraints = [qt.encode_label(self.con, subject_type, label, merge=False) for label in labels]
        label_constraints = [i for i in label_constraints if i is not None]

        n = len(label_constraints)

        if n == 0:
            # None of the requested labels exists yet, so there certainly aren't any objects having them.
            raise KnownEmptyResult

        label_constraint = label_constraints[0] if n == 1 else label_constraints

        if subject_type == qt.SubjectTypes.VERTICES:
            query = qt.write_quads_query(
                sel='s', s=s_subquery, p=qt.Constants.HAS_LABEL_PREDICATE, o=label_constraint
            )
        elif subject_type == qt.SubjectTypes.EDGES:
            query = qt.write_quads_query(
                sel='g', p=label_constraint, g=s_subquery
            )
        else:
            raise BadStepArgs(filter_step.err_msg('Unknown subject type'))  # pragma: no cover

        return FilterableQuery(subject_type, query)



# --------------------------------------------------------------------------- #
#   Copyright (c) 2024-2026 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

"""
Implements a simple adapter so that the Gremlin query language can be used with Python's sqlite3
module, for a serverless, file-based graph database.
"""

import pathlib
import sqlite3
import sys
import uuid

from gremlin_python.driver.remote_connection import RemoteTraversal
from gremlin_python.process.anonymous_traversal import traversal
from gremlin_python.process.traversal import Bytecode

from .base import GremliteConfig, ProducerInternalResultIterator
from .errors import BadDatabase
from .logging import OpenCloseLoggingConnection, PlanLoggingConnection
import gremlite.querytools as qt
from .steps import bytecode_to_producer_chain


# At least due to our use of keyword "RETURNING" (it is used in many places
# in the `querytools.py` module), we need SQLite 3.35 or later.
REQUIRED_SQLITE_MINIMUM_VERSION = (3, 35)


class InsufficientSQLiteVersionException(Exception):

    def __init__(self, detected_version, extra_phrase=None):
        self.detected_version = detected_version
        self.extra_phrase = extra_phrase or ' '

    def __str__(self):
        major, minor = REQUIRED_SQLITE_MINIMUM_VERSION
        msg = f'GremLite requires SQLite {major}.{minor}.x or later'
        msg += f', but version {self.detected_version}{self.extra_phrase}was detected.'
        return msg


def check_sqlite_version(config: GremliteConfig):
    """
    Check that we are using an acceptable version of SQLite.

    :param config: this is used for testing purposes
    :return: nothing
    :raises: InsufficientSQLiteVersionException
    """
    version_string = sqlite3.sqlite_version

    if config.phony_testing_sqlite_version_string is not None:
        version_string = config.phony_testing_sqlite_version_string

    parts = version_string.split('.')

    # If version not in three parts, we don't know how to read it:
    if len(parts) != 3:
        raise InsufficientSQLiteVersionException(
            version_string, extra_phrase=' of unrecognized format '
        )

    # I hope at least major and minor are always integers, but just in case:
    try:
        major, minor = [int(p) for p in parts[:2]]
    except ValueError:
        raise InsufficientSQLiteVersionException(
            version_string, extra_phrase=' with non-integer component(s) '
        )

    if (major, minor) < REQUIRED_SQLITE_MINIMUM_VERSION:
        raise InsufficientSQLiteVersionException(version_string)


def get_g(db_file_path,
          autocommit=None, isolation_level="DEFERRED", timeout=5.0,
          log_plans=False, session=None, config: GremliteConfig = None):
    """
    Convenience function to get a Gremlin graph traversal source for a given
    SQLite database file.

    :return: GraphTraversalSource
    """
    remote = SQLiteConnection(
        db_file_path,
        autocommit=autocommit, isolation_level=isolation_level, timeout=timeout,
        log_plans=log_plans, session=session, config=config
    )
    g = traversal().with_remote(remote)
    return g


class SQLiteConnection:
    """
    Stand-in for the `DriverRemoteConnection` class of gremlinpython.

    Whereas ordinary usage of gremlinpython to connect to an actual Gremlin server
    might look like this:

        uri = 'ws://localhost:8182/gremlin'
        remote = DriverRemoteConnection(uri)
        g = traversal().with_remote(remote)

    usage of this `SQLiteConnection` class instead looks like this:

        path = '/filesystem/path/to/my_sqlite_database_file.db'
        remote = SQLiteConnection(path)
        g = traversal().with_remote(remote)

    References
    ==========

    gremlinpython:
        https://pypi.org/project/gremlinpython/

    Gremlin:
        https://tinkerpop.apache.org/gremlin.html
        https://tinkerpop.apache.org/docs/current/reference/
        https://arxiv.org/abs/1508.03843

    SQLite:
        https://sqlite.org
        https://docs.python.org/3/library/sqlite3.html

    """

    def __init__(self, db_file_path,
                 autocommit=None, isolation_level="DEFERRED", timeout=5.0,
                 log_plans=False, check_qqc_patterns=False, log_open_close=0,
                 session=None, config: GremliteConfig = None):
        """
        :param db_file_path: string or pathlib.Path pointing to the file you want
            to use for your on-disk database with SQLite. The file can be a pre-existing one, from a
            previous use of GremLite, or else should not yet exist (for an initial use of GremLite).
            In all cases, the directory named by the path must already exist.
        :param autocommit: Set to a boolean value to control transactions as in Python 3.12 and forward.
            See https://docs.python.org/3.12/library/sqlite3.html#transaction-control
            To be clear, you can use the ``autocommit`` kwarg here even if you are using a version of
            Python prior to 3.12. This library interprets it for you.
        :param isolation_level: If not setting the ``autocommit`` kwarg, use this to control transactions
            as in Python versions prior to 3.12.
            See https://docs.python.org/3.8/library/sqlite3.html#controlling-transactions
            At time of writing, Python versions 3.12 and later also continue to support ``isolation_level``.
        :param timeout: number of seconds to wait to acquire transaction locks (default 5.0 s)
        :param log_plans: set True to record query plans via INFO-level logging
        :param check_qqc_patterns: set True to raise an exception if any unexpected "quad query constraint"
            pattern is used. This is mainly for development and diagnostics.
        :param log_open_close: set to 1 to keep track of open cursors and make their stack traces inspectible
            via gremlite.logging.print_open_cursor_traces(). Set to 2 to also log open/close events via INFO-level
            logging. This param does nothing if `log_plans` is True.
        :param session: used internally to start transactions
        :param config: optionally pass a `GremliteConfig` to make settings.
        """
        self.db_file_path = pathlib.Path(db_file_path)

        self.is_logging = log_plans is not False
        if check_qqc_patterns:
            qt.check_qqc_patterns()

        self.__session = session
        self.config = config or GremliteConfig()
        check_sqlite_version(self.config)

        connection_constructor = sqlite3.Connection
        if log_plans:
            connection_constructor = PlanLoggingConnection
        elif log_open_close > 0:
            connection_constructor = lambda database, timeout=5.0: OpenCloseLoggingConnection(
                database, timeout=timeout, log_level=log_open_close)

        self.con = connection_constructor(self.db_file_path, timeout=timeout)

        """
        The meaning of the Connection's `isolation_level` property is a bit subtle.
        To understand it, we have to first be clear that SQLite *itself* has its own
        behavior, while the Python sqlite3 *module* has *its* own behavior, and these
        are two different things.
        
        The default behavior of SQLite, independent of Python, is autocommit mode. This means
        that *if you have not explicitly said BEGIN to start a transaction* then any
        statement automatically initiates a transaction,
        *and* this transaction is automatically committed as soon as the statement is
        completed. See:
            https://sqlite.org/lang_transaction.html
            https://system.data.sqlite.org/index.html/raw/419906d4a0043afe2290c2be186556295c25c724
        
        On the contrary, the default behavior you will see if you use SQLite via Python's
        sqlite3 module is different. Like SQLite's built-in behavior, a transaction is
        automatically *begun* for any DML statement (INSERT/UPDATE/DELETE/REPLACE)
        if you have not already explicitly
        started one earlier using `BEGIN`; however, *unlike* SQLite's built-in behavior,
        this transaction is *not* automatically *committed*. So: begun yes, committed no.
        You can carry out a sequence of DML statements, and their effects will pile up in
        the transaction that Python silently started for you, but none of these effects will
        be committed until you explicitly call the `commit()` method on your `Connection`.
        See https://docs.python.org/3.8/library/sqlite3.html#controlling-transactions
        
        If you accept the default values `autocommit=None, isolation_level="DEFERRED"` of this
        `SQLiteConnection.__init__()` kwargs, you are going to get the Python sqlite3 module's
        default behavior.
        
        To recover SQLite's native autocommit behavior, you can set the `Connection` object's
        `isolation_level` attribute to `None`. Or, in Python 3.12 or later, it now has an
        `autocommit` attribute, which you can set to `True` (or you can continue to use the
        `isolation_level`, which is still there too).
        """

        if isinstance(autocommit, bool):
            vi = sys.version_info
            if vi.major >= 3 and vi.minor >= 12:
                self.con.autocommit = autocommit  # pragma: no cover
            else:
                self.con.isolation_level = None if autocommit else "DEFERRED"
        else:
            self.con.isolation_level = isolation_level

        if self.is_session_bound():
            cur = self.con.cursor()
            cur.execute('BEGIN TRANSACTION;')
            cur.close()

        if not db_already_initialized(self.con):
            initialize_db(self.con)

        self.__spawned_sessions = []

    def table_stats(self):
        return qt.get_table_stats(self.con)

    def close(self):
        if len(self.__spawned_sessions) > 0:
            for spawned_session in self.__spawned_sessions:
                spawned_session.close()
            self.__spawned_sessions.clear()

        self.con.close()

    def submit(self, bytecode: Bytecode) -> RemoteTraversal:
        """
        As in the `DriverRemoteConnection` class of gremlinpython, the `submit()`
        method is the one that is invoked through `Traversal.__next__()` when it
        calls `apply_strategies()`. This is where it is our job to take the bytecode
        representing a Gremlin traversal, and construct an iterator producing the
        results of the query.
        """
        producer = bytecode_to_producer_chain(self.config, self.con, bytecode)

        # This is purely for testing purposes:
        if self.config.traversal_returns_internal_result_iterator:
            producer = ProducerInternalResultIterator(producer)

        return RemoteTraversal(producer)

    def is_session_bound(self):
        return self.__session is not None

    def create_session(self):
        """
        Supports the `gremlin_python.process.graph_traversal.Transaction.begin()` method.
        """
        if self.is_session_bound():
            raise Exception(
                'Connection is already bound to a session - child sessions are not allowed')  # pragma: no cover
        con = SQLiteConnection(self.db_file_path, log_plans=self.is_logging, session=uuid.uuid4())
        self.__spawned_sessions.append(con)
        return con

    def remove_session(self, session_based_connection):
        session_based_connection.close()
        self.__spawned_sessions.remove(session_based_connection)

    def commit(self):
        self.con.commit()

    def rollback(self):
        self.con.rollback()


# These are the indexes that we use, on the quads table:
quad_index_names = """
ops
sp
pg
g
""".split()


def initialize_db(con: sqlite3.Connection):
    """
    Create the tables and indexes that we need.

    Data Model
    ==========

    We use a variant of the "RDF quads" design for recording a property graph in a 4-column table in a relational
    database, as described in

        Harth, Andreas, and Stefan Decker. "Optimized index structures for
        querying RDF from the web." In Third Latin American Web Congress
        (LA-WEB'2005), pp. 10-pp. IEEE, 2005.

        https://ieeexplore.ieee.org/abstract/document/1592360

    See also:

        https://docs.aws.amazon.com/neptune/latest/userguide/feature-overview-data-model.html

    Indexes
    -------

    We use a different set of indexes than those suggested by either of the two above references, namely,
    we use the four indexes: ops, sp, pg, g.

    To begin with, while those sources consider each column's constraints to be binary (on or off), for us they are
    ternary: exact (equal to a particular value or in a finite set of values), sign (pos. or neg.), or none. This is
    important because in order for SQLite to make use of an index, the columns with exact constraints have to come
    before those with sign constraints. As stated on this page: https://www.sqlite.org/optoverview.html,
    "The initial columns of the index must be used with the = or IN or IS operators. The right-most column that is
    used can employ inequalities."

    We describe a query to the quads table by a pattern like "[po|s]", which means that we are putting exact
    constraints on the p and o columns, a sign constraint on the s column, and no constraint on the g column.

    At the time of developing our first release of this project, we performed a survey of all the quad constraint
    patterns that were actually used in all of the Gremlin steps we were supporting at that time. (Our unit tests
    record them in a log, and we have code to survey the log.) This resulted in the list of patterns recorded in
    the `expected_quad_constraint_patterns` in the `querytools.py` module.

                                                [|g]        (g)
                                                [s|p]       (sp)
                                                [s|]        (sp)
                                                [sp|]       (sp)
                                                [spo|]      (ops)
                                                [p|g]       (pg)
                                                [p|]        (pg)
                                                [po|s]      (ops)
                                                [po|]       (ops)
                                                [pg|]       (pg)
                                                [o|p]       (ops)
                                                [g|]        (g)

                            The expected query patterns, and the indexes that cover them.

    After studying this list we chose a set of four indexes that cover all these patterns. (For an index to cover
    a pattern [L|R] it has to be possible to choose a permutation of L, and a permutation of R, and concatenate these,
    and thereby get some prefix of the index description.) Note that, among our expected patterns, each of the four
    column letters s, p, o, g occurs alone on the left-hand side of at least one pattern, which means that we need
    *at least* four indexes: one to support each column as the sole exact constraint. Therefore it seems plausible
    that our chosen list of four indexes might be optimal (though I can't say for sure).

    Our unit tests contain checks (a) that no unexpected quad constraint patterns are used, and (b) that all
    expected ones are used. If check (a) should fail as we move forward and support more of the Gremlin
    language, it may mean that our new code should be reworked to stay within the set of expected patterns, or
    it may mean that we need a new index. But the latter should be avoided if possible, since indexes impose both space
    costs (file size) and time costs when writing to the database. If check (b) should fail, it would mean that we
    had already reworked things somehow, and should reevaluate whether we still have the best set of indexes.

    IDs, Valence, and Property Value Encodings
    ------------------------------------------

    ID numbers are assigned to three types of "graph elements:" vertices, edges, and vertex properties.
    (Edge properties are not treated the same way. Everything we do is modeled on the behavior of TinkerGraph 3.7.2.)
    These three types of graph elements share a single ID space of positive integers. Thus, no two of these will
    ever have the same ID. ID numbers are never reused after objects are deleted from the graph.

    ID numbers are also assigned to strings, which encompass vertex and edge labels, property names, and string or
    float property values. These too are auto-incrementing positive integers, but they start at 3 (in order to
    make room for encoding property values of other types, including boolean and null).

    When IDs are recorded in the table of quads, they are sometimes negated, as indicated below, in order to resolve
    any possible ambiguities. We refer to the sign of an ID when recorded in the quads table as its "valence."
    To be clear, all IDs themselves are positive integers; they are only sometimes *recorded* with negative sign.

        For the S column ("subject"):

            edges: < 0
            vertices: > 0

        For the P column ("predicate"):

            edge labels: < 0
            "has label": = 0
            property names: > 0

        For the O column ("object"):

            vertex labels: > 0

            vertices: > 0

            integer: negative (add 2^62 to recover actual value)
            boolean: 0 or 1 for False or True resp.
            null: 2
            strings & floats: >= 3. When greater than 2^62, interpreted as float.

        For the G column (edge ID or vertex property ID):

            edges: < 0
            vertex properties: > 0

    Integer property values must lie in the range [-2^62, 2^62 - 1]. They are shifted by -2^62
    before being recorded (thus, are always negative in the table).

    The "O" column is the only one in which different types share certain integer ranges, but
    the value in the "P" column always disambiguates it:

        * O can only be "vertices" if P < 0 ("edge labels")
        * O can only be "vertex labels" if P = 0 ("has label")
        * O can only be "int/bool/null/string/float" if P > 0 ("property names")

    Note that column "O" never holds edge labels because those are recorded differently;
    namely, they go in the "P" column in quads where the "G" column holds the (negated) edge ID.

    Quad Types
    ----------

    There are only four different types of quads in the quads table, which correspond to
    the four different types of things that there are in a property graph: vertices, edges,
    vertex properties, and edge properties.

    From the encoding described above you can figure out for yourself what the four quad types
    look like, but we summarize them here for convenience.

    Vertex:

        (+vertex ID, 0, +vertex label ID, 0)

    Edge:

        (+source vertex ID, -edge label ID, +target vertex ID, -edge ID)

    Vertex Property:

        (+vertex ID, +property name ID, encoded property value, +vertex property ID)

    Edge Property:

        (-edge ID, +property name ID, encoded property value, 0)

    It follows that each quad type can be recognized immediately by a single column condition:

        Vertex: p = 0
        Edge:   p < 0 or g < 0 (either condition on its own is sufficient)
        Prop:   p > 0 (this identifies a property of any kind, vertex or edge)
        V Prop: g > 0
        E Prop: s < 0

    """
    cur = con.cursor()

    cur.execute("CREATE TABLE quads (s INT, p INT, o INT, g INT)")

    for name in quad_index_names:
        cur.execute(f"CREATE INDEX {name} ON quads ({', '.join(list(name))})")

    cur.execute("CREATE TABLE strings (id INTEGER PRIMARY KEY AUTOINCREMENT, v TEXT UNIQUE)")
    # Note: There is no need to create an index to look up string IDs by string values,
    # since SQLite will just use its own internal index `sqlite_autoindex_strings_1` for that anyway,
    # which it forms in order to enforce the UNIQUE constraint.
    # See https://stackoverflow.com/a/35626972

    # Ensure that string IDs start at 3:
    cur.execute("INSERT INTO strings (v) VALUES ('foo')")
    cur.execute("INSERT INTO strings (v) VALUES ('bar')")
    cur.execute("DELETE FROM strings WHERE id = 1")
    cur.execute("DELETE FROM strings WHERE id = 2")
    # In fact 3 will always be the ID of the empty string. We want ID 3 to actually be
    # present, so that we can make a positive check in the `db_already_initialized()` function.
    cur.execute("INSERT INTO strings (v) VALUES ('')")

    # For graph elements (vertices, edges, and vertex properties) there is nothing to record except their ID.
    # So this table just serves to record the set of existing IDs, as well as to produce new IDs.
    # We use an AUTOINCREMENT id column instead of built-in rowid, to ensure
    # that ids are never reused. See https://stackoverflow.com/a/9342301
    cur.execute("CREATE TABLE graphelts (id INTEGER PRIMARY KEY AUTOINCREMENT)")

    cur.close()
    con.commit()


expected_table_names = {'quads', 'strings', 'graphelts'}

# We include a check that sqlite's 'sqlite_autoindex_strings_1' is formed, because we are relying on that.
expected_index_names = set(quad_index_names).union({'sqlite_autoindex_strings_1'})


def db_already_initialized(con: sqlite3.Connection):
    """
    Check if a database has been initialized, with exactly the tables
    and indexes that we expect.

    :param con: connection
    :return: boolean, True if it looks like the database has already been initialized,
        False if it looks brand new.
    :raise: BadDatabase if there are one or more tables or indexes, but they are not
        as we expect. This should only happen if you connected to some database
        used for some other purpose than ours.
    """
    cur = con.cursor()
    try:
        cur.execute("SELECT name FROM sqlite_master WHERE type='table'")
        actual_table_names = {r[0] for r in cur.fetchall()}

        cur.execute("SELECT name FROM sqlite_master WHERE type='index'")
        actual_index_names = {r[0] for r in cur.fetchall()}

        # Among the actual tables and indices, there will be special ones like 'sqlite_sequence'.
        # We don't want to worry about those, so we just check that all expected ones are present.
        if expected_table_names.issubset(actual_table_names) and expected_index_names.issubset(actual_index_names):

            # Check that string IDs start at 3
            R = cur.execute("SELECT rowid FROM strings WHERE rowid <= 3").fetchall()
            if len(R) != 1 or R[0][0] != 3:
                raise BadDatabase

            return True
        elif len(actual_table_names) == 0 and len(actual_index_names) == 0:
            return False
        else:
            raise BadDatabase
    finally:
        cur.close()



# --------------------------------------------------------------------------- #
#   Copyright (c) 2024-2026 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

"""
Custom connection and cursor classes for logging purposes
"""

import logging
import os
import sqlite3
import time
import traceback
import uuid

OPEN_CURSORS = {}


def generate_new_uid(prefix=''):
    ts = time.time_ns()
    return f'{prefix}-{ts}-{uuid.uuid4()}'


class PlanLoggingConnection(sqlite3.Connection):
    """
    This connection class, and the associated cursor class, serve to log query plans,
    each time a query is made.
    """

    def cursor(self):
        sqlite_cursor = super().cursor()
        return PlanLoggingCursor(sqlite_cursor)


class PlanLoggingCursor:
    """
    Cursor formed by PlanLoggingConnection, to log query plans.
    """

    def __init__(self, sqlite_cursor):
        self.sqlite_cursor = sqlite_cursor
        self.logger = logging.getLogger(__name__)

    def execute(self, sql, params=None):
        params = params or []

        explain_sql = 'EXPLAIN QUERY PLAN ' + sql
        self.sqlite_cursor.execute(explain_sql, params)
        strat = self.sqlite_cursor.fetchall()

        self.logger.info(sql)
        self.logger.info(strat)

        return self.sqlite_cursor.execute(sql, params)

    def fetchone(self):
        return self.sqlite_cursor.fetchone()

    def fetchall(self):
        return self.sqlite_cursor.fetchall()

    def close(self):
        return self.sqlite_cursor.close()


class OpenCloseLoggingConnection(sqlite3.Connection):
    """
    This connection class, and the associated cursor class, serve to log opening and closing
    of connections and cursors. Can help to track down issues with multiple open cursors
    causing "database locked" errors.
    """

    def __init__(self, database, timeout=5.0, log_level=1):
        super().__init__(database, timeout=timeout)
        self.gremlite_uid = generate_new_uid(prefix='CONN')
        self.log_level = log_level
        self.logger = logging.getLogger(__name__)
        note_pid = f'PID={os.getpid()}'
        self.log('~~~~~~ OP', extra=note_pid)

    def log(self, action, extra=''):
        if self.log_level < 2:
            return
        ts = time.time_ns()
        msg = f'{action} ({ts}) {self.gremlite_uid} {extra}'
        self.logger.info(msg)

    def cursor(self):
        sqlite_cursor = super().cursor()
        return OpenCloseLoggingCursor(self, sqlite_cursor, log_level=self.log_level)

    def commit(self) -> None:
        self.log('CMT')
        return super().commit()

    def close(self):
        self.log('~~~~~~ CL')
        return super().close()


class OpenCloseLoggingCursor:
    """
    Cursor formed by OpenCloseLoggingConnection, to log open/close events.

    log_level:
        1: Just record self (with stack trace) in the `OPEN_CURSORS` registry.
        2: Also record open/close events with INFO-level logging.
    """

    def __init__(self, connection, sqlite_cursor, log_level=1):
        self.gremlite_uid = generate_new_uid(prefix='CURS')
        self.log_level = log_level

        if self.log_level >= 1:
            OPEN_CURSORS[self.gremlite_uid] = traceback.format_stack()

        self.connection = connection
        self.sqlite_cursor = sqlite_cursor

        self.logger = logging.getLogger(__name__)
        self.log('OPN')

    def log(self, action, extra=''):
        if self.log_level < 2:
            return
        ts = time.time_ns()
        msg = f'{action} ({ts}) {self.gremlite_uid} {self.connection.gremlite_uid} {extra}'
        self.logger.info(msg)

    def execute(self, sql, params=None):
        params = params or []
        self.log('EXC', extra=sql)
        return self.sqlite_cursor.execute(sql, params)

    def fetchone(self):
        return self.sqlite_cursor.fetchone()

    def fetchall(self):
        return self.sqlite_cursor.fetchall()

    def close(self):
        result = self.sqlite_cursor.close()
        self.log('CLS')
        if self.log_level >= 1:
            del OPEN_CURSORS[self.gremlite_uid]
        return result


def print_open_cursor_traces(limit=5):
    """
    Convenience function to print a representation of all currently open cursors,
    when using an OpenCloseLoggingConnection.

    :param limit: upper limit on how many open cursors' full stack trace will be printed
    """
    N = len(OPEN_CURSORS)
    print(f'There are {N} open cursors.')
    for uid, trace in list(OPEN_CURSORS.items())[:min(N, limit)]:
        print("%" * 80)
        print(uid)
        for line in trace:
            print(line.rstrip('\n'))



# --------------------------------------------------------------------------- #
#   Copyright (c) 2024-2026 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from .connection import get_g, SQLiteConnection, GremliteConfig

__all__ = [
    'get_g', 'SQLiteConnection', 'GremliteConfig',
]

__version__ = '0.37.0-rc1'



# --------------------------------------------------------------------------- #
#   Copyright (c) 2024-2026 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from enum import Enum
import logging
import sqlite3

import gremlin_python.process.traversal as predicates

from .cursors import DummyCursor
from .errors import UnexpectedQuadConstraintPattern


logger = logging.getLogger(__name__)


class Constants:
    MIN_INT = -2**62
    MAX_INT = 2**62 - 1
    HAS_LABEL_PREDICATE = 0
    IMPOSSIBLE_SUBJECT = 0
    NON_ID_G_VALUE = 0


class SubjectTypes(Enum):
    VERTICES = 'VERTICES'
    EDGES = 'EDGES'


class Sign(Enum):
    NEG = -1
    POS = 1

    @classmethod
    def valence(cls, st: SubjectTypes):
        return cls.NEG if st is SubjectTypes.EDGES else cls.POS


def take_next_graphelt_id(con: sqlite3.Connection):
    cur = con.cursor()
    next_id = cur.execute("INSERT INTO graphelts VALUES (NULL) RETURNING id").fetchone()[0]
    cur.close()
    return next_id


def create_vertex(con: sqlite3.Connection, vertex_label: str):
    """
    Fully create a new vertex. Affects both the graphelts table and the quads table.

    :param con: connection
    :param vertex_label: the label for the new vertex
    :return: the ID assigned to the new vertex
    """
    vertex_id = take_next_graphelt_id(con)
    label_id = merge_string(con, vertex_label)
    cur = con.cursor()
    cur.execute("INSERT INTO quads VALUES (?, ?, ?, ?)", [
        vertex_id, Constants.HAS_LABEL_PREDICATE, label_id, Constants.NON_ID_G_VALUE
    ])
    cur.close()
    return vertex_id


def create_edge(con: sqlite3.Connection, source_id: int, edge_label: str, target_id: int):
    """
    Fully create a new edge. Affects both the graphelts table and the quads table.

    :param con: connection
    :param source_id: id of source vertex
    :param edge_label: the label for the new edge
    :param target_id: id of target vertex
    :return: the ID assigned to the new edge
    """
    edge_id = take_next_graphelt_id(con)
    label_id = merge_string(con, edge_label)
    cur = con.cursor()
    cur.execute("INSERT INTO quads VALUES (?, ?, ?, ?)", [source_id, -label_id, target_id, -edge_id])
    cur.close()
    return edge_id


def create_vertex_property(con: sqlite3.Connection, vertex_id: int, prop_name: str, prop_value):
    """
    Fully create a new vertex property. Affects both the graphelts table and the quads table.

    :param con: connection
    :param vertex_id: the id of the vertex
    :param prop_name: the name of the property
    :param prop_value: the value of the property
    :return: the ID assigned to the new vertex property
    """
    prop_id = take_next_graphelt_id(con)
    name_id = merge_string(con, prop_name)
    value_code = encode_property_value(con, prop_value, merge=True)
    cur = con.cursor()
    cur.execute("INSERT INTO quads VALUES (?, ?, ?, ?)", [
        vertex_id, name_id, value_code, prop_id
    ])
    cur.close()
    return prop_id


def record_edge_property(con: sqlite3.Connection, edge_id: int, prop_name: str, prop_value):
    """
    Record an edge property in the quads table. Since there are no edge property graph elements,
    we do not form a new element ID.

    :param con: connection
    :param edge_id: the id of the edge
    :param prop_name: the name of the property
    :param prop_value: the value of the property
    :return: nothing
    """
    name_id = merge_string(con, prop_name)
    value_code = encode_property_value(con, prop_value, merge=True)
    cur = con.cursor()
    cur.execute("INSERT INTO quads VALUES (?, ?, ?, ?)", [
        -edge_id, name_id, value_code, Constants.NON_ID_G_VALUE
    ])
    cur.close()


def replace_vertex_property(con: sqlite3.Connection, old_prop_id, new_prop_value):
    """
    Replace an existing vertex property.

    Removes the existing vertex property as a graph element,
    forms a new graph element, and updates the existing quad row to
    record the new property value and ID.

    :param con: connection
    :param old_prop_id: the existing property ID
    :param new_prop_value: the new value for the property
    :return: nothing
    """
    delete_elements_from_graphelts_table(con, old_prop_id)
    new_prop_id = take_next_graphelt_id(con)
    value_code = encode_property_value(con, new_prop_value, merge=True)
    cur = con.cursor()
    log_qqc('g')
    cur.execute("UPDATE quads SET o = ?, g = ? WHERE g = ?", [value_code, new_prop_id, old_prop_id])
    cur.close()


def update_edge_property(con: sqlite3.Connection, edge_id, prop_name, new_prop_value):
    """
    Update the quads table for a new value of an existing edge property.

    :param con: connection
    :param edge_id: the edge ID
    :param prop_name: the name of the property
    :param new_prop_value: the new value for the property
    :return:
    """
    name_id = string_to_id(con, prop_name)
    value_code = encode_property_value(con, new_prop_value, merge=True)
    cur = con.cursor()
    log_qqc('sp')
    cur.execute("UPDATE quads SET o = ? WHERE s = ? AND p = ?", [value_code, -edge_id, name_id])
    cur.close()


def delete_elements_from_graphelts_table(con: sqlite3.Connection, element_ids):
    """
    Delete one or more elements from the graphelts table.

    This does NOT perform all the necessary actions to completely remove an element
    from the graph. E.g. removing a vertex also requires removing all of its properties
    and connected edges.

    Instead, this is just one step in a complete removal of an element.

    :param con: connection
    :param element_ids: either a single int, or a list of ints
    :return: number of IDs actually deleted
    """
    if isinstance(element_ids, int):
        element_ids = [element_ids]

    id_set = ','.join(str(n) for n in element_ids)
    cur = con.cursor()
    ids_deleted = cur.execute(
        f"DELETE FROM graphelts WHERE id IN ({id_set}) RETURNING id"
    ).fetchall()
    cur.close()

    return len(ids_deleted)


def get_table_stats(con: sqlite3.Connection):
    """
    Get stats on what is in the database.

    :param con: connection
    :return: RowCount
    """
    cur = con.cursor()

    # Check how many we have of each type of object:
    # Vertices:
    vx = cur.execute('SELECT count(*) FROM quads WHERE p = 0').fetchone()[0]
    # Edges:
    eg = cur.execute('SELECT count(*) FROM quads WHERE g < 0').fetchone()[0]
    # Vertex properties:
    vp = cur.execute('SELECT count(*) FROM quads WHERE g > 0').fetchone()[0]
    # Edge properties:
    ep = cur.execute('SELECT count(*) FROM quads WHERE s < 0').fetchone()[0]

    # Sanity checks:
    num_quads = cur.execute('SELECT count(*) FROM quads').fetchone()[0]
    num_graph_elts = cur.execute('SELECT count(*) FROM graphelts').fetchone()[0]
    assert vx + eg + vp + ep == num_quads
    assert vx + eg + vp == num_graph_elts

    # Strings:
    st = cur.execute('SELECT count(*) FROM strings').fetchone()[0]

    cur.close()

    rc = RowCount()
    rc.qvl = vx
    rc.qe = eg
    rc.qvp = vp
    rc.qep = ep
    rc.s = st
    return rc


class RowCount:
    """Records counts of row types"""

    def __init__(self):
        # quads table, edge rows (p < 0)
        self.qe = 0
        # quads table, vertex label rows (p = 0)
        self.qvl = 0
        # quads table, vertex property rows (p > 0, s > 0)
        self.qvp = 0
        # quads table, edge property rows (p > 0, s < 0)
        self.qep = 0

        # graphelts table, vertex rows
        self.gv = 0
        # graphelts table, edge rows
        self.ge = 0
        # graphelts table, vertex property rows
        self.gvp = 0

        # strings table rows
        self.s = 0

    def __add__(self, other):
        rc = RowCount()

        rc.qe = self.qe + other.qe
        rc.qvl = self.qvl + other.qvl
        rc.qvp = self.qvp + other.qvp
        rc.qep = self.qep + other.qep

        rc.gv = self.gv + other.gv
        rc.ge = self.ge + other.ge
        rc.gvp = self.gvp + other.gvp

        rc.s = self.s + other.s

        return rc

    def __str__(self):
        parts = []
        for a in "qe qvl qvp qep gv ge gvp s".split():
            parts.append(f'{a}: {getattr(self, a)}')
        return f'({", ".join(parts)})'

    def matches(self, tup):
        """
        Pass a tuple of integers. We say whether our counts, in canonical order
        (i.e. order initialized in __init__() method), equal these, resp.

        :param tup: tuple of ints
        :return: boolean
        """
        return all(
            getattr(self, a) == tup[i]
            for i, a in enumerate("qe qvl qvp qep gv ge gvp s".split())
        )


def completely_remove_vertex(con: sqlite3.Connection, vertex_id):
    """
    Completely remove a vertex.

    This completely removes the vertex, as well as all incident edges, and all
    properties of the vertex and its incident edges. All necessary rows from both
    the quads table and the graphelts table are removed.

    :param con: connection
    :param vertex_id: the ID of the vertex to be removed
    :return: RowCount
    """
    rc = RowCount()

    # Vertex label:
    log_qqc('sp')
    cur = con.cursor()
    label_codes = cur.execute(
        "DELETE FROM quads WHERE s = ? AND p = 0 RETURNING o",
        [vertex_id]
    ).fetchall()
    cur.close()
    rc.qvl = len(label_codes)

    # Vertex ID:
    rc.gv = delete_elements_from_graphelts_table(con, vertex_id)

    # Vertex property quads:
    log_qqc('s', 'p')
    cur = con.cursor()
    rows = cur.execute(
        "DELETE FROM quads WHERE s = ? AND p > 0 RETURNING g",
        [vertex_id]
    ).fetchall()
    cur.close()
    vprop_ids = [r[0] for r in rows]
    rc.qvp = len(vprop_ids)

    # Vertex property IDs:
    rc.gvp = delete_elements_from_graphelts_table(con, vprop_ids)

    # Outgoing edges:
    log_qqc('s', 'p')
    cur = con.cursor()
    rows = cur.execute(
        "SELECT g FROM quads WHERE s = ? AND p < 0",
        [vertex_id]
    ).fetchall()
    cur.close()
    negative_edge_ids = [r[0] for r in rows]
    rc += completely_remove_edges(con, negative_edge_ids)

    # Incoming edges:
    log_qqc('o', 'p')
    cur = con.cursor()
    rows = cur.execute(
        "SELECT g FROM quads WHERE o = ? AND p < 0",
        [vertex_id]
    ).fetchall()
    cur.close()
    negative_edge_ids = [r[0] for r in rows]
    rc += completely_remove_edges(con, negative_edge_ids)

    return rc


def completely_remove_edges(con: sqlite3.Connection, negative_edge_ids):
    """
    Completely remove a set of edges.

    This completely removes the edges, as well as all of their properties.
    All necessary rows from both the quads table and the graphelts table are removed.

    :param con: connection
    :param negative_edge_ids: int or list of ints, being the *negative valence* IDs
        of the edges to be deleted. (I.e. you should be passing negative integers.)
    :return: RowCount
    """
    if isinstance(negative_edge_ids, int):
        negative_edge_ids = [negative_edge_ids]

    rc = RowCount()
    cur = con.cursor()

    # Edge quads:
    id_set = ','.join(str(n) for n in negative_edge_ids)
    log_qqc('g')
    rows = cur.execute(
        f"DELETE FROM quads WHERE g IN ({id_set}) RETURNING -g"
    ).fetchall()
    positive_ids_deleted = [r[0] for r in rows]
    rc.qe = len(positive_ids_deleted)

    # Edge property quads:
    log_qqc('s', 'p')
    eprops = cur.execute(
        f"DELETE FROM quads WHERE s IN ({id_set}) AND p > 0 RETURNING p"
    ).fetchall()
    rc.qep = len(eprops)

    cur.close()

    # Edge IDs:
    rc.ge = delete_elements_from_graphelts_table(con, positive_ids_deleted)

    return rc


def completely_remove_vertex_property(con: sqlite3.Connection, vprop_id):
    """
    Completely remove a vertex property.

    This removes a row from both the quads table and the graphelts table.

    :param con: connection
    :param vprop_id: the ID of the vertex property to be removed.
    :return: RowCount
    """
    rc = RowCount()

    cur = con.cursor()
    log_qqc('g')
    ids_deleted = cur.execute(
        "DELETE FROM quads WHERE g = ? RETURNING g", [vprop_id]
    ).fetchall()
    cur.close()
    rc.qvp = len(ids_deleted)

    rc.gvp = delete_elements_from_graphelts_table(con, vprop_id)

    return rc


def completely_remove_edge_property(con: sqlite3.Connection, edge_id, prop_name):
    """
    Completely remove an edge property.

    :param con: connection
    :param edge_id: the ID of the edge (pos or neg valence is okay) whose property
        is to be removed.
    :param prop_name: string, being the name of the property that is to be removed.
    :return: RowCount
    """
    rc = RowCount()

    if edge_id > 0:
        edge_id = -edge_id

    cur = con.cursor()
    log_qqc('sp')
    eprops = cur.execute(
        "DELETE FROM quads WHERE s = ? AND p IN (SELECT id FROM strings WHERE v = ?) RETURNING p",
        [edge_id, prop_name]
    ).fetchall()
    cur.close()
    rc.qep = len(eprops)

    return rc


def merge_string(con: sqlite3.Connection, val):
    """
    Retrieve the ID of an existing string, or add a new string and return its
    newly formed ID.

    :param con: connection
    :param val: string value
    :return: string ID
    """
    cur = con.cursor()
    r = cur.execute("SELECT id FROM strings WHERE v = ?", [val]).fetchone()
    if r is None:
        # The string is not yet present
        cur.execute("INSERT INTO strings (v) VALUES (?) RETURNING id", [val])
        sid = cur.fetchone()[0]
    else:
        # The string is already present.
        sid = r[0]
    cur.close()
    return sid


def string_to_id(con: sqlite3.Connection, val):
    """
    Retrieve the ID of a string.

    :param con: connection
    :param val: the string
    :return: int or None
    """
    cur = con.cursor()
    cur.execute("SELECT id FROM strings WHERE v = ?", [val])
    r = cur.fetchone()
    cur.close()
    return r if r is None else r[0]


def id_to_string(con: sqlite3.Connection, i):
    """
    Retrieve the string of a given ID.

    :param con: connection
    :param i: the ID
    :return: string or None
    """
    cur = con.cursor()
    cur.execute('SELECT v FROM strings WHERE id = ?', [i])
    r = cur.fetchone()
    cur.close()
    return r if r is None else r[0]


def get_string_id(con: sqlite3.Connection, val, merge=False):
    if merge:
        return merge_string(con, val)
    else:
        return string_to_id(con, val)


def encode_property_name(con: sqlite3.Connection, name, merge=False):
    """
    Turn a property name into an int that is ready to be recorded in the quads table.

    :param con: connection
    :param name: the property name
    :param merge: set True to add the name if not already present
    :return: int or None
    """
    return get_string_id(con, name, merge=merge)


def encode_property_value(con: sqlite3.Connection, value, merge=False):
    """
    Turn a property value into a constraint for the o column in the quads table.

    :param con: connection
    :param value: None, boolean, int in the range [-2^62, 2^62 - 1], float, string, or an instance of
        one of gremlinpython's `P` or `TextP` predicate classes.
    :param merge: only affects string values. Set True to add the string if not already present.
    :return: int or None
    """
    if value is False:
        return 0
    elif value is True:
        return 1
    elif value is None:
        return 2
    elif isinstance(value, int):
        if value < Constants.MIN_INT or value > Constants.MAX_INT:
            return None
        else:
            return value + Constants.MIN_INT
    elif isinstance(value, float):
        str_value = str(value)
        base_id = get_string_id(con, str_value, merge=merge)
        return None if base_id is None else base_id + Constants.MAX_INT
    elif isinstance(value, str):
        return get_string_id(con, value, merge=merge)
    elif isinstance(value, predicates.P):
        return get_predicate_column_constraint(con, value, ColumnRole.PROPERTY_VALUE)
    return None


def decode_property_value(con: sqlite3.Connection, o):
    """
    Given the integer representing a property value in the quads table o column,
    return the actual property value.

    :param con: connection
    :param o: int representing prop value
    :return: prop value
    """
    if o == 0:
        return False
    elif o == 1:
        return True
    elif o == 2:
        return None
    elif o < 0:
        return o - Constants.MIN_INT
    elif o > Constants.MAX_INT:
        base_id = o - Constants.MAX_INT
        str_value = id_to_string(con, base_id)
        return float(str_value)
    else:
        return id_to_string(con, o)


def encode_vertex_label(con: sqlite3.Connection, value, merge=False):
    if isinstance(value, str):
        return get_string_id(con, value, merge=merge)
    elif isinstance(value, predicates.P):
        return get_predicate_column_constraint(con, value, ColumnRole.VERTEX_LABEL)
    return None


def encode_edge_label(con: sqlite3.Connection, value, merge=False):
    if isinstance(value, str):
        i = get_string_id(con, value, merge=merge)
        return None if i is None else -i
    elif isinstance(value, predicates.P):
        return get_predicate_column_constraint(con, value, ColumnRole.EDGE_LABEL)
    return None


def encode_label(con: sqlite3.Connection, subject_type: SubjectTypes, value, merge=False):
    if subject_type == SubjectTypes.VERTICES:
        return encode_vertex_label(con, value, merge=merge)
    elif subject_type == SubjectTypes.EDGES:
        return encode_edge_label(con, value, merge=merge)


def decode_vertex_label(con: sqlite3.Connection, o_col_value):
    return id_to_string(con, o_col_value)


def decode_edge_label(con: sqlite3.Connection, p_col_value):
    return id_to_string(con, -p_col_value)


def get_vertex_label(con: sqlite3.Connection, vertex_id):
    log_qqc('sp')
    cur = con.cursor()
    label = cur.execute("SELECT v FROM strings WHERE id IN (SELECT o FROM quads WHERE s = ? AND p = ?)", [
        vertex_id, Constants.HAS_LABEL_PREDICATE
    ]).fetchone()[0]
    cur.close()
    return label


def get_edge_label(con: sqlite3.Connection, edge_id):
    log_qqc('g')
    cur = con.cursor()
    label = cur.execute("SELECT v FROM strings WHERE id IN (SELECT -p FROM quads WHERE g = ?)", [
        -edge_id
    ]).fetchone()[0]
    cur.close()
    return label


def get_label(con: sqlite3.Connection, obj_id: int, subject_type: SubjectTypes):
    if subject_type == SubjectTypes.VERTICES:
        return get_vertex_label(con, obj_id)
    elif subject_type == SubjectTypes.EDGES:
        return get_edge_label(con, obj_id)


class InterpolatedQuery(str):
    """
    Use this to represent query strings containing one or more question marks, and keep
    their list of parameters associated with them. Supports left and right addition with
    regular strings or other InterpolatedQuery instances, concatenating the parameter
    lists appropriately.
    """

    def __new__(cls, value, params=None):
        self = str.__new__(cls, value)
        self.value = value
        self.params = params or []
        return self

    def __str__(self):
        return f'({self.value}, {self.params})'  # pragma: no cover

    def __add__(self, other):
        if isinstance(other, InterpolatedQuery):
            return InterpolatedQuery(self.value + other.value, self.params + other.params)
        if isinstance(other, str):
            iq = InterpolatedQuery(other)
            return self + iq
        raise NotImplemented  # pragma: no cover

    def __radd__(self, other):
        if isinstance(other, str):
            iq = InterpolatedQuery(other)
            return iq + self
        raise NotImplemented  # pragma: no cover


def query_join(glue, queries):
    """
    Perform a string join in such a way that InterpolatedQuery's will work properly.

    :param glue: the string to connect the queries
    :param queries: list of queries (str or InterpolatedQuery instances)
    :return: str or InterpolatedQuery
    """
    if not queries:
        return ''
    result = queries[0]
    for q in queries[1:]:
        result += glue + q
    return result


def write_strings_to_ids_query(strings, id_valence=1):
    """
    Given a finite list of strings, write a query string selecting the IDs of those strings.

    :param strings: nonempty list of desired strings
    :param id_valence: set to -1 to select negated IDs
    :return: string
    """
    sel = '-id' if id_valence == -1 else 'id'
    qms = ','.join(['?']*len(strings))
    return InterpolatedQuery(f"SELECT {sel} FROM strings WHERE v IN ({qms})", strings)


class ColumnRole(Enum):
    EDGE = 'EDGE'
    VERTEX = 'VERTEX'

    EDGE_LABEL = 'EDGE_LABEL'
    VERTEX_LABEL = 'VERTEX_LABEL'

    PROPERTY_NAME = 'PROPERTY_NAME'
    PROPERTY_VALUE = 'PROPERTY_VALUE'


def is_text_predicate(obj):
    """
    Say whether a given object is a text predicate.

    Since `TextP.gt` etc. actually return instances of `P`, not `TextP`, we
    need a slightly more sophisticated check in order to determine if we are
    looking at a text predicate.

    :param obj: any object
    :return: boolean
    """
    return isinstance(obj, predicates.TextP) or (
            isinstance(obj, predicates.P) and
            isinstance(obj.value, str) and
            obj.operator in ['gt', 'lt', 'gte', 'lte']
    )


class PredicateHandler:

    def handle_predicate(self, pred: predicates.P):
        operator = pred.operator
        value = pred.value

        if is_text_predicate(pred):
            if operator == 'startingWith':
                return self.starting_with(value)
            elif operator == 'containing':
                return self.containing(value)
            elif operator == 'endingWith':
                return self.ending_with(value)
            elif operator == 'gt':
                return self.gt(value)
            elif operator == 'lt':
                return self.lt(value)
            elif operator == 'gte':
                return self.gte(value)
            elif operator == 'lte':
                return self.lte(value)

        elif isinstance(pred, predicates.P):
            if operator == 'within':
                return self.within(value)

    def starting_with(self, value):
        raise NotImplementedError  # pragma: no cover

    def containing(self, value):
        raise NotImplementedError  # pragma: no cover

    def ending_with(self, value):
        raise NotImplementedError  # pragma: no cover

    def gt(self, value):
        raise NotImplementedError  # pragma: no cover

    def lt(self, value):
        raise NotImplementedError  # pragma: no cover

    def gte(self, value):
        raise NotImplementedError  # pragma: no cover

    def lte(self, value):
        raise NotImplementedError  # pragma: no cover

    def within(self, value):
        raise NotImplementedError  # pragma: no cover


class PredicateColumnConstraintHandler(PredicateHandler):

    def __init__(self, con: sqlite3.Connection, column_role: ColumnRole):
        self.con = con
        self.column_role = column_role

        text_sign = '-' if column_role is ColumnRole.EDGE_LABEL else ''
        self.text_p_basic = f"SELECT {text_sign}id FROM strings WHERE v "

    def build_interp_text_query(self, ending, value):
        return InterpolatedQuery(self.text_p_basic + ending, [value])

    def starting_with(self, value):
        return self.build_interp_text_query("GLOB ?", f'{value}*')

    def containing(self, value):
        return self.build_interp_text_query("GLOB ?", f'*{value}*')

    def ending_with(self, value):
        return self.build_interp_text_query("GLOB ?", f'*{value}')

    def gt(self, value):
        return self.build_interp_text_query("> ?", value)

    def lt(self, value):
        return self.build_interp_text_query("< ?", value)

    def gte(self, value):
        return self.build_interp_text_query(">= ?", value)

    def lte(self, value):
        return self.build_interp_text_query("<= ?", value)

    def within(self, value):
        values = value
        con = self.con
        column_role = self.column_role

        if column_role == ColumnRole.EDGE:
            # We assume you passed a list of edge IDs, and at worst you
            # forgot to make them negative, but you did use consistent sign throughout.
            if len(values) > 0 and values[0] > 0:
                return [-v for v in values]
            else:
                return values
        elif column_role == ColumnRole.VERTEX:
            # We assume you passed a list of vertex IDs, and they're all positive.
            # No change required.
            return values

        # For the following cases, it's possible that `None` be returned when we attempt to
        # encode, due to strings not being present in the strings table. So we need to filter
        # those cases before returning.
        if column_role == ColumnRole.EDGE_LABEL:
            # We assume you passed a list of strings.
            values = [encode_edge_label(con, v) for v in values]
        elif column_role == ColumnRole.VERTEX_LABEL:
            # We assume you passed a list of strings.
            values = [encode_vertex_label(con, v) for v in values]
        elif column_role == ColumnRole.PROPERTY_NAME:
            # We assume you passed a list of strings.
            values = [encode_property_name(con, v) for v in values]
        elif column_role == ColumnRole.PROPERTY_VALUE:
            # We assume you passed a list of types that can be encoded as property values
            # (so None, bool, int, float, or str).
            values = [encode_property_value(con, v) for v in values]

        return [v for v in values if v is not None]


def get_predicate_column_constraint(con: sqlite3.Connection, pred: predicates.P, column_role: ColumnRole):
    """
    Generate a constraint that can be passed to the `get_quads()` function.

    :param con: connection
    :param pred: an instance of gremlinpython's `P` or `TextP` predicate classes
    :param column_role: a value of the ColumnRole enum class, indicating the
        role this predicate is playing
    :return: constraint that can be passed to the `get_quads()` function.
    """
    handler = PredicateColumnConstraintHandler(con, column_role)
    return handler.handle_predicate(pred)


def write_quads_query(sel="*", s=None, p=None, o=None, g=None):
    """
    Write an sqlite query against the quads table

    :param sel: what you want to SELECT from the quads table
    :param s: Constrain the s component of the quad:
        * None: no constraint
        * Sign.NEG or Sign.POS: require this sign
        * a single integer: require that it equal this value
        * a list of integers: require that it be in this set
        * string: require that it be in the set returned by this string as a subquery
    :param p: like s
    :param o: like s
    :param g: like s
    :return: query string
    """
    given = {'s': s, 'p': p, 'o': o, 'g': g}

    primary = []
    secondary = []

    conditions = {}
    for k, v in given.items():
        if v is not None:
            if v is Sign.NEG:
                conditions[k] = f'{k} < 0'
                secondary.append(k)
            elif v is Sign.POS:
                conditions[k] = f'{k} > 0'
                secondary.append(k)
            elif isinstance(v, int):
                conditions[k] = f'{k} = {v}'
                primary.append(k)
            elif isinstance(v, list):
                conditions[k] = f'{k} IN ({",".join(str(n) for n in v)})'
                primary.append(k)
            elif isinstance(v, str):
                # Use explicit `+` here in case v is an InterpolatedQuery:
                conditions[k] = f'{k} IN (' + v + ')'
                primary.append(k)

    log_qqc(primary, secondary)

    condition_seq = list(conditions.values())

    where_clause = ''
    if condition_seq:
        where_clause = ' WHERE ' + query_join(' AND ', condition_seq)

    query = f"SELECT {sel} FROM quads" + where_clause

    return query


def run_quads_query(con: sqlite3.Connection, sel="*", s=None, p=None, o=None, g=None, do_close=False, fetch_one=False):
    query = write_quads_query(sel, s, p, o, g)
    return run_query(con, query, do_close=do_close, fetch_one=fetch_one)


def run_query(con: sqlite3.Connection, query, do_close=False, fetch_one=False):
    cur = con.cursor()
    if isinstance(query, InterpolatedQuery):
        cur.execute(query.value, query.params)
    else:
        cur.execute(query)
    if do_close:
        rows = [cur.fetchone()] if fetch_one else cur.fetchall()
        cur.close()
        return DummyCursor(rows)
    else:
        return cur


# We build a dictionary mapping all known expected quad constraint patterns to the
# index we expect to be used for each.
_ix_pairs = """
[|g]        (g)
[s|p]       (sp)
[s|]        (sp)
[sp|]       (sp)
[spo|]      (ops)
[p|g]       (pg)
[p|]        (pg)
[po|s]      (ops)
[po|]       (ops)
[pg|]       (pg)
[o|p]       (ops)
[g|]        (g)
""".split()
expected_quad_constraint_patterns = dict(zip(_ix_pairs[0::2], _ix_pairs[1::2]))


qqc_header = 'QUAD QUERY CONSTRAINTS: '


def format_qqc(primary, secondary=''):
    """
    Write a string of the form "[...|...]" to describe quad query constraints.

    The letters in each half will be sorted for you, so you can pass them in any order.

    :param primary: string or list of column names receiving exact constraints
    :param secondary: optional string or list of column names receiving inexact constraints
    :return: string
    """
    return f'[{"".join(sorted(primary, reverse=True))}|{"".join(sorted(secondary, reverse=True))}]'


def log_qqc(primary, secondary=''):
    """
    Log a message noting quad query constraints.

    :param primary: string or list of column names
    :param secondary: optional string or list of column names
    :return: nothing
    """
    qqc = format_qqc(primary, secondary)
    msg = qqc_header + qqc
    logger.info(msg)


class QqcPatternFilter(logging.Filter):
    """
    Raise exception on presence of unexpected QQC pattern.
    """

    def filter(self, record: logging.LogRecord) -> bool:
        t = record.msg
        if t.startswith(qqc_header):
            i0 = len(qqc_header)
            i1 = t.index(']')
            pattern = t[i0:i1+1]
            if pattern not in expected_quad_constraint_patterns:
                raise UnexpectedQuadConstraintPattern(pattern)
        return False


qqc_handler = logging.StreamHandler()
qqc_handler.addFilter(QqcPatternFilter())


def check_qqc_patterns():
    """
    Invoke if you want exceptions to be raised on unexpected QQC patterns.
    """
    logger.addHandler(qqc_handler)
